#ifndef _NETACUITY_API_H_
#define _NETACUITY_API_H_
/*
* ***************************************************************************
* File:           netacuity_api.h
* Author:         Digital Envoy
* Program Name:   Personum API library
* Date:           2/25/00
* Version:        3.0
*
*  Description:
*    These functions allow a developer to access the NetAcuity Database.
*    These routines are MT-Safe
*
*  Revision History:
*  ----------------------------
*   $Log: netacuity_api.h,v $
*   Revision 1.1  2004/01/17 00:26:39  bhuffake
*
*       Added code that will allows sk_analysis_dump to talk to netacuity_api.
*
*   Revision 1.1  2003/10/16 19:15:47  bhuffake
*
*   	Adding netacq to the cvs repository.
*
*   Revision 1.2  2002/07/02 13:55:10  jnystrom
*   Made additions for new Databases
*
*   Revision 1.1  2002/05/14 14:51:26  tmcneese
*   Move APIs to individual directories.
*
*   Revision 3.3  2002/03/23 00:10:00  jnystrom
*   Put version number in standard header.
*
*   Revision 3.2  2002/03/15 23:52:11  jnystrom
*   Changed the return codes on the new functions.
*
*   Revision 3.1  2002/03/01 23:03:25  jnystrom
*   Made function for domain and sic db querying and got rid of "personum" references
*
*   Added new functions for querying the domain and sic code db as well
*   as the geo db.  The references to "personum" were also done away with.
*  ----------------------------
* **************************************************************************
*/

/* DB feature codes */
#define NA_GEO      3
#define NA_SIC      5
#define NA_DOMAIN   6 
#define NA_ZIP      7
#define NA_ISP      8
#define NA_HOME_BIZ 9


/* Database response structure */
typedef struct {
    char IP[16];       // IP address as a string in decimal dotted notation
                       // Currently IP is NOT filled in by API (for future)
    // These fields will be filled in by the API on successfull call to
    // the Personum_query() and na_query_geo()
    char country[4];   // Country code returned from database
    char region[32];   // region (state in the US)
    char city[32];     // city name
    char speed[11];    // connection speed (this is a character string value)
    int country_c;     // confidence percentage that the country is correct
    int region_c;      // region confidence
    int city_c;        // city confidence
    int metro_code;    // metro_code
    float latitude;    // latitude
    float longitude;   // longitude
} na_geo_struct;

/* struct for SIC */
typedef struct {
    char IP[16];          // currently not used
    int  sic;             // SIC
} na_sic_struct;

/* Domain struct */
typedef struct {
    char IP[16];          // currently not used
    char domain[256];     // domain name
} na_domain_struct;

/* struct for zip/area code/timezone DB */
typedef struct{
    char IP[16];          // currently not used
    int  area_code;       // area code
    int  zip;             // zip code
    int  current_offset;  // current offset from UTC (GMT)
    char in_dst[2];       // in daytime savings time or not
}na_zip_struct;

/* struct for isp */
typedef struct{
    char IP[16];       //currently not used
    char isp[236];     //the isp
}na_isp_struct;

typedef struct{
    char IP[16];       //currently not used
    char home_biz[236]; //home or busisness
}na_home_biz_struct;
    
/* API Functions */

/***
   Function:    na_api_set_id
   Returns:     void
   Parameters:  Integer between 1 and 127.  0 is left as "unregistered"
   Description: Initializes the client's key into the library.  This function
                must be called before a query can proceed sucessfully.
***/
void na_api_set_id(int id);

/***
   Function:    na_api_set_server_addr
   Returns:     zero=error in ip address.  nonzero=success
   Parameters:  ip address string in decimal dotted notation
   Description: Initializes the NetAcuity server to query for data
                must be called before a query can proceed sucessfully.
***/
int na_api_set_server_addr(char *addr);

/***
   Function:    na_api_set_timeout
   Returns:     void
   Parameters:  timout time in microseconds (us)
   Description: Sets the maximum amount of time to wait on network calls
                before giving up.  (this includes both connect and query)
                The default timeout value is 250ms (250000 us)
***/
void na_api_set_timeout(int t);

/***
   Function:    na_query_raw
   Returns:     1-success, 0=error
   Parameters:  A character dot-notation value for the IP address to query.
                The pointer for the return response, integer pointer for
                the number of fields in the response, pointer to an
                int for the response size to be returned in of the response,
                db_feature.
   Description: queries the NetAcuity database specified in db_feature for the
                specified IP address and returns to size of the respone, the
                number of fields in the response, and the response itself.  If
                there is a timeout or error, it returns 0 and the size is set
                to -1, num_fields is set to 0, and response is not filled in.
***/
int na_query_raw(char *ip, char *response, int *num_fields, int *size,
                 int db_feature);

/***
   Function:    na_query_geo
   Returns:     1-success, 0-error
   Parameters:  A character dot notation value for the IP address to query
                and a pointer to an na_geo_struct, which will be filled in
                with the response from the NetAcuity server.
   Description: queries the NA geo database with the specified IP address
                and returns the geo information in the na_geo_struct.  The size
                of the response includes the '\n' included.
***/
int na_query_geo(char *ip, na_geo_struct *r);

/***
   Function:    na_query_domain
   Returns:     1-success, 0-error
   Parameters:  A character dotted notation of the IP address to query
                and a pointer to an na_domain_struct that will be filled in by
                the NetAcuity api.
   Description: queries the NA domain db with the specified IP address and
                returns the information in the struct passed in.
***/
int na_query_domain(char *ip, na_domain_struct *r);

/***
   Function:    na_query_sic
   Returns:     1-success, 0-error
   Parameters:  A character dotted notation of the IP address to query and
                a pointer to an na_sic_struct that will be filled in by the
                NetAcuity api.
   Description: queries the NA sic db with the specified IP address and
                returns the information in the struct passed in.

***/

int na_query_sic(char *ip, na_sic_struct *r);

/***
   Function:    na_query_zip
   Returns:     0-success, -1-error
   Parameters:  A character dotted notation of the IP address to query.
                The na_zip_struct that will be filled in by the netacuity
                api.
   Description: This function queries the NA zip/area code/timezone database
                and returns the information in the struct passed to the
                function.
***/

int na_query_zip(char *ip, na_zip_struct *r);

/***
   Function:    na_query_isp
   Returns:     0-success, -1-error
   Parameters:  A character dotted notation of the IP address to query.
                The na_isp_struct that will be filled in by the netacuity
                api.
   Description: This function queries the NA ISP database
                and returns the information in the struct passed to the
                function.
***/

int na_query_isp(char *ip, na_isp_struct *r);

/***
   Function:    na_query_home_biz
   Returns:     0-success, -1-error
   Parameters:  A character dotted notation of the IP address to query.
                The na_home_biz_struct that will be filled in by the netacuity
                api.
   Description: This function queries the NA home/business database
                and returns the information in the struct passed to the
                function.
***/

int na_query_home_biz(char *ip, na_home_biz_struct *r);

#endif
